from typing import Annotated

from pydantic import Field

from phylogenie.core.dataset import DatasetGenerator
from phylogenie.core.msas import MSAsGeneratorConfig
from phylogenie.core.trees import TreesGeneratorConfig

DatasetGeneratorConfig = Annotated[
    TreesGeneratorConfig | MSAsGeneratorConfig,
    Field(discriminator="data_type"),
]

__all__ = ["DatasetGeneratorConfig", "DatasetGenerator"]



from typing import Annotated

from pydantic import Field

from phylogenie.core.msas.alisim import AliSimGenerator

MSAsGeneratorConfig = Annotated[
    AliSimGenerator,
    Field(discriminator="backend"),
]



import subprocess
from typing import Literal

from numpy.random import Generator

import phylogenie.typings as pgt
from phylogenie.core.msas.base import BackendType, MSAsGenerator


class AliSimGenerator(MSAsGenerator):
    backend: Literal[BackendType.ALISIM] = BackendType.ALISIM
    iqtree_path: str = "iqtree2"
    args: dict[str, str | int | float]

    def _generate_one_from_tree(
        self, filename: str, tree_file: str, rng: Generator, data: pgt.Data
    ) -> None:
        command = [
            self.iqtree_path,
            "--alisim",
            filename,
            "--tree",
            tree_file,
            "--seed",
            str(rng.integers(0, 2**32 - 1)),
        ]

        for key, value in self.args.items():
            command.extend(
                [key, value.format(**data) if isinstance(value, str) else str(value)]
            )

        command.extend(["-af", "fasta"])
        subprocess.run(command, check=True, stdout=subprocess.DEVNULL)
        subprocess.run(["rm", f"{tree_file}.log"], check=True)



from collections.abc import Callable, Iterator
from typing import TypeGuard, Union

import phylogenie.typeguards as tg
import phylogenie.typings as pgt
from phylogenie.skyline.parameter import (
    SkylineParameter,
    SkylineParameterLike,
    is_skyline_parameter_like,
    skyline_parameter,
)
from phylogenie.skyline.vector import (
    SkylineVector,
    SkylineVectorLike,
    SkylineVectorOperand,
    is_many_skyline_vectors_like,
    is_skyline_vector_like,
    is_skyline_vector_operand,
    skyline_vector,
)

SkylineMatrixOperand = Union[SkylineVectorOperand, "SkylineMatrix"]
SkylineMatrixCoercible = Union[
    SkylineParameterLike, pgt.Many[SkylineVectorLike], "SkylineMatrix"
]


def is_skyline_matrix_operand(x: object) -> TypeGuard[SkylineMatrixOperand]:
    return isinstance(x, SkylineMatrix) or is_skyline_vector_operand(x)


class SkylineMatrix:
    def __init__(
        self,
        params: pgt.Many[SkylineVectorLike] | None = None,
        value: pgt.Many3DScalars | None = None,
        change_times: pgt.ManyScalars | None = None,
    ):
        if params is not None and value is None and change_times is None:
            if is_many_skyline_vectors_like(params):
                self.params = [skyline_vector(p, len(params)) for p in params]
            else:
                raise TypeError(
                    f"It is impossible to create a SkylineMatrix from `params` {params} of type {type(params)}. Please provide a sequence composed of SkylineVectorLike objects (a SkylineVectorLike object can either be a SkylineVector or a sequence of scalars and/or SkylineParameters)."
                )
        elif value is not None and change_times is not None:
            if tg.is_many_3D_scalars(value):
                matrix_lengths = {len(matrix) for matrix in value}
                if any(ml != len(value[0]) for ml in matrix_lengths):
                    raise ValueError(
                        f"All matrices in the `value` of a SkylineMatrix must have the same length (got value={value} with matrix lengths={matrix_lengths})."
                    )
            else:
                raise TypeError(
                    f"It is impossible to create a SkylineMatrix from `value` {value} of type {type(value)}. Please provide a nested (3D) sequence of scalar values."
                )
            self.params = [
                SkylineVector(
                    value=[matrix[i] for matrix in value], change_times=change_times
                )
                for i in range(len(value[0]))
            ]
        else:
            raise ValueError(
                "Either `params` or both `value` and `change_times` must be provided to create a SkylineMatrix."
            )

    @property
    def N(self) -> int:
        return len(self.params)

    @property
    def change_times(self) -> pgt.Vector1D:
        return sorted(set([t for row in self.params for t in row.change_times]))

    @property
    def value(self) -> pgt.Vector3D:
        return [self.get_value_at_time(t) for t in (0, *self.change_times)]

    def get_value_at_time(self, time: pgt.Scalar) -> pgt.Vector2D:
        return [param.get_value_at_time(time) for param in self.params]

    def operate(
        self,
        other: SkylineMatrixOperand,
        func: Callable[
            [SkylineVector, SkylineVector | SkylineParameter], SkylineVector
        ],
    ) -> "SkylineMatrix":
        if not is_skyline_matrix_operand(other):
            return NotImplemented
        if is_skyline_vector_operand(other):
            other = skyline_vector(other, N=self.N)
        assert isinstance(other, SkylineVector | SkylineMatrix)
        return SkylineMatrix(
            [func(p1, p2) for p1, p2 in zip(self.params, other.params)]
        )

    def __add__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: x + y)

    def __radd__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: y + x)

    def __sub__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: x - y)

    def __rsub__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: y - x)

    def __mul__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: x * y)

    def __rmul__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: y * x)

    def __truediv__(self, operand: SkylineMatrixOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: x / y)

    def __rtruediv__(self, operand: SkylineVectorOperand) -> "SkylineMatrix":
        return self.operate(operand, lambda x, y: y / x)

    @property
    def T(self) -> "SkylineMatrix":
        return SkylineMatrix([[v[i] for v in self] for i in range(self.N)])

    def __bool__(self) -> bool:
        return any(self.params)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SkylineMatrix) and self.params == other.params

    def __repr__(self) -> str:
        return f"SkylineMatrix(value={list(self.value)}, change_times={list(self.change_times)})"

    def __iter__(self) -> Iterator[SkylineVector]:
        return iter(self.params)

    def __len__(self) -> int:
        return self.N

    def __getitem__(self, item: int) -> "SkylineVector":
        return self.params[item]

    def __setitem__(self, item: int, value: SkylineVectorLike) -> None:
        if not is_skyline_vector_like(value):
            raise TypeError(
                f"It is impossible to set item {item} of SkylineMatrix with value {value} of type {type(value)}. Please provide a SkylineVectorLike object (i.e., a SkylineVector or a sequence of scalars and/or SkylineParameters)."
            )
        self.params[item] = skyline_vector(value, N=self.N)


def skyline_matrix(
    x: SkylineMatrixCoercible, N: int, zero_diagonal: bool = False
) -> SkylineMatrix:
    if is_skyline_parameter_like(x):
        x = SkylineMatrix([[skyline_parameter(x)] * N] * N)
        if zero_diagonal:
            for i in range(N):
                x[i][i] = 0
        return x
    elif is_many_skyline_vectors_like(x):
        x = SkylineMatrix(x)
    if not isinstance(x, SkylineMatrix):
        raise TypeError(
            f"It is impossible to coerce {x} of type {type(x)} into a SkylineMatrix. Please provide either:\n"
            "- a SkylineMatrix,\n"
            "- a SkylineParameterLike object (i.e., a scalar or a SkylineParameter),\n"
            "- a sequence of SkylineVectorLike objects (a SkylineVectorLike object can be a SkylineVector or a sequence of SkylineParameterLike objects)."
        )

    if x.N != N:
        raise ValueError(
            f"Expected an {N}x{N} SkylineMatrix, got {x} of shape {x.N}x{x.N}."
        )

    if zero_diagonal and any(x[i][i] for i in range(x.N)):
        raise ValueError(f"Expected a SkylineMatrix with zero diagonal, but got {x}.")

    return x



from typing import Annotated

from pydantic import Field

from phylogenie.core.trees.remaster import ReMASTERGeneratorConfig
from phylogenie.core.trees.treesimulator import TreeSimulatorGeneratorConfig

TreesGeneratorConfig = Annotated[
    ReMASTERGeneratorConfig | TreeSimulatorGeneratorConfig,
    Field(discriminator="backend"),
]



from enum import Enum
from typing import Annotated, Literal

import numpy as np
from numpy.random import Generator
from pydantic import Field

import phylogenie.core.configs as cfg
import phylogenie.typings as pgt
from phylogenie.backend.treesimulator import (
    DEFAULT_POPULATION,
    TreeParams,
    generate_tree,
    get_BD_params,
    get_BDEI_params,
)
from phylogenie.core.factories import (
    int_factory,
    scalar_factory,
    skyline_matrix_coercible_factory,
    skyline_parameter_like_factory,
    skyline_vector_coercible_factory,
)
from phylogenie.core.trees.base import BackendType, TreesGenerator


class ParameterizationType(str, Enum):
    CANONICAL = "canonical"
    BD = "BD"
    BDEI = "BDEI"


class TreeSimulatorGenerator(TreesGenerator):
    backend: Literal[BackendType.TREESIMULATOR] = BackendType.TREESIMULATOR
    min_tips: cfg.IntConfig
    max_tips: cfg.IntConfig
    T: cfg.ScalarConfig = np.inf
    root_state: str | None = None
    state_frequencies: list[float] | None = None
    notification_probability: cfg.SkylineParameterLikeConfig = 0
    notification_sampling_rate: cfg.SkylineParameterLikeConfig = np.inf
    allow_irremovable_states: bool = False
    max_notified_contacts: cfg.IntConfig = 1

    def _generate_one_from_params(
        self, filename: str, rng: Generator, data: pgt.Data, params: TreeParams
    ) -> None:
        root_state = (
            self.root_state
            if self.root_state is None
            else self.root_state.format(**data)
        )
        generate_tree(
            output_file=f"{filename}.nwk",
            params=params,
            min_tips=int_factory(self.min_tips, data),
            max_tips=int_factory(self.max_tips, data),
            T=scalar_factory(self.T, data),
            state_frequencies=self.state_frequencies,
            notification_probability=skyline_parameter_like_factory(
                self.notification_probability, data
            ),
            notification_sampling_rate=skyline_parameter_like_factory(
                self.notification_sampling_rate, data
            ),
            allow_irremovable_states=self.allow_irremovable_states,
            max_notified_contacts=int_factory(self.max_notified_contacts, data),
            root_state=root_state,
            random_seed=int(rng.integers(0, 2**31 - 1)),
        )


class CanonicalTreeSimulatorGenerator(TreeSimulatorGenerator):
    parameterization: Literal[ParameterizationType.CANONICAL] = (
        ParameterizationType.CANONICAL
    )
    populations: str | list[str] = DEFAULT_POPULATION
    transition_rates: cfg.SkylineMatrixCoercibleConfig = 0
    transmission_rates: cfg.SkylineMatrixCoercibleConfig = 0
    removal_rates: cfg.SkylineVectorCoercibleConfig = 0
    sampling_proportions: cfg.SkylineVectorCoercibleConfig = 0

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        self._generate_one_from_params(
            filename,
            rng,
            data,
            TreeParams(
                populations=self.populations,
                transition_rates=skyline_matrix_coercible_factory(
                    self.transition_rates, data
                ),
                transmission_rates=skyline_matrix_coercible_factory(
                    self.transmission_rates, data
                ),
                removal_rates=skyline_vector_coercible_factory(
                    self.removal_rates, data
                ),
                sampling_proportions=skyline_vector_coercible_factory(
                    self.sampling_proportions, data
                ),
            ),
        )


class BDTreeSimulatorGenerator(TreeSimulatorGenerator):
    parameterization: Literal[ParameterizationType.BD] = ParameterizationType.BD
    reproduction_number: cfg.SkylineParameterLikeConfig = 0
    infectious_period: cfg.SkylineParameterLikeConfig = 0
    sampling_proportion: cfg.SkylineParameterLikeConfig = 0

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        self._generate_one_from_params(
            filename,
            rng,
            data,
            get_BD_params(
                reproduction_number=skyline_parameter_like_factory(
                    self.reproduction_number, data
                ),
                infectious_period=skyline_parameter_like_factory(
                    self.infectious_period, data
                ),
                sampling_proportion=skyline_parameter_like_factory(
                    self.sampling_proportion, data
                ),
            ),
        )


class BDEITreeSimulatorGenerator(TreeSimulatorGenerator):
    parameterization: Literal[ParameterizationType.BDEI] = ParameterizationType.BDEI
    reproduction_number: cfg.SkylineParameterLikeConfig = 0
    infectious_period: cfg.SkylineParameterLikeConfig = 0
    incubation_period: cfg.SkylineParameterLikeConfig = 0
    sampling_proportion: cfg.SkylineParameterLikeConfig = 0

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        self._generate_one_from_params(
            filename,
            rng,
            data,
            get_BDEI_params(
                reproduction_number=skyline_parameter_like_factory(
                    self.reproduction_number, data
                ),
                infectious_period=skyline_parameter_like_factory(
                    self.infectious_period, data
                ),
                incubation_period=skyline_parameter_like_factory(
                    self.incubation_period, data
                ),
                sampling_proportion=skyline_parameter_like_factory(
                    self.sampling_proportion, data
                ),
            ),
        )


TreeSimulatorGeneratorConfig = Annotated[
    CanonicalTreeSimulatorGenerator
    | BDTreeSimulatorGenerator
    | BDEITreeSimulatorGenerator,
    Field(discriminator="parameterization"),
]



from dataclasses import dataclass

import phylogenie.typings as pgt
from phylogenie.skyline import (
    SkylineMatrixCoercible,
    SkylineParameterLike,
    SkylineVectorCoercible,
    skyline_matrix,
    skyline_vector,
)

SAMPLE_POPULATION = "sample"
DEFAULT_POPULATION = "X"


@dataclass
class Reaction:
    rate: SkylineParameterLike
    value: str


@dataclass
class PunctualReaction:
    times: pgt.ManyScalars
    value: str
    p: pgt.ManyScalars | None = None
    n: pgt.Many[int] | None = None


def get_canonical_reactions(
    populations: str | list[str] = DEFAULT_POPULATION,
    sample_population: str = SAMPLE_POPULATION,
    birth_rates: SkylineVectorCoercible = 0,
    death_rates: SkylineVectorCoercible = 0,
    sampling_rates: SkylineVectorCoercible = 0,
    removal_probabilities: SkylineVectorCoercible = 0,
    migration_rates: SkylineMatrixCoercible = 0,
    birth_rates_among_demes: SkylineMatrixCoercible = 0,
) -> list[Reaction]:
    if isinstance(populations, str):
        populations = [populations]
    N = len(populations)

    birth_rates = skyline_vector(birth_rates, N=N)
    death_rates = skyline_vector(death_rates, N=N)
    sampling_rates = skyline_vector(sampling_rates, N=N)
    removal_probabilities = skyline_vector(removal_probabilities, N=N)
    migration_rates = skyline_matrix(migration_rates, N=N, zero_diagonal=True)
    birth_rates_among_demes = skyline_matrix(
        birth_rates_among_demes, N=N, zero_diagonal=True
    )

    reactions: list[Reaction] = []
    for i, population in enumerate(populations):
        reactions.append(Reaction(birth_rates[i], f"{population} -> 2{population}"))
        reactions.append(Reaction(death_rates[i], f"{population} -> 0"))
        reactions.append(
            Reaction(
                sampling_rates[i] * removal_probabilities[i],
                f"{population} -> {sample_population}",
            )
        )
        reactions.append(
            Reaction(
                sampling_rates[i] * (1 - removal_probabilities[i]),
                f"{population} -> {population} + {sample_population}",
            )
        )
        for j, other_population in enumerate(populations):
            if i == j:
                continue
            reactions.append(
                Reaction(migration_rates[i][j], f"{population} -> {other_population}")
            )
            reactions.append(
                Reaction(
                    birth_rates_among_demes[i][j],
                    f"{population} -> {population} + {other_population}",
                )
            )
    return reactions


def get_epidemiological_reactions(
    populations: str | list[str] = DEFAULT_POPULATION,
    sample_population: str = SAMPLE_POPULATION,
    reproduction_numbers: SkylineVectorCoercible = 0,
    become_uninfectious_rates: SkylineVectorCoercible = 0,
    sampling_proportions: SkylineVectorCoercible = 0,
    removal_probabilities: SkylineVectorCoercible = 0,
    migration_rates: SkylineMatrixCoercible = 0,
    reproduction_numbers_among_demes: SkylineMatrixCoercible = 0,
) -> list[Reaction]:
    if isinstance(populations, str):
        populations = [populations]
    N = len(populations)

    reproduction_numbers = skyline_vector(reproduction_numbers, N=N)
    become_uninfectious_rates = skyline_vector(become_uninfectious_rates, N=N)
    sampling_proportions = skyline_vector(sampling_proportions, N=N)
    removal_probabilities = skyline_vector(removal_probabilities, N=N)
    reproduction_numbers_among_demes = skyline_matrix(
        reproduction_numbers_among_demes, N=N, zero_diagonal=True
    )

    birth_rates = reproduction_numbers * become_uninfectious_rates
    birth_rates_among_demes = (
        reproduction_numbers_among_demes * become_uninfectious_rates
    )
    sampling_rates = become_uninfectious_rates * sampling_proportions
    death_rates = become_uninfectious_rates - removal_probabilities * sampling_rates

    return get_canonical_reactions(
        populations=populations,
        sample_population=sample_population,
        birth_rates=birth_rates,
        death_rates=death_rates,
        sampling_rates=sampling_rates,
        removal_probabilities=removal_probabilities,
        migration_rates=migration_rates,
        birth_rates_among_demes=birth_rates_among_demes,
    )


def get_FBD_reactions(
    populations: str | list[str] = DEFAULT_POPULATION,
    sample_population: str = SAMPLE_POPULATION,
    diversification: SkylineVectorCoercible = 0,
    turnover: SkylineVectorCoercible = 0,
    sampling_proportions: SkylineVectorCoercible = 0,
    removal_probabilities: SkylineVectorCoercible = 0,
    migration_rates: SkylineMatrixCoercible = 0,
    diversification_between_types: SkylineMatrixCoercible = 0,
):
    if isinstance(populations, str):
        populations = [populations]
    N = len(populations)

    diversification = skyline_vector(diversification, N=N)
    turnover = skyline_vector(turnover, N=N)
    sampling_proportions = skyline_vector(sampling_proportions, N=N)
    removal_probabilities = skyline_vector(removal_probabilities, N=N)
    diversification_between_types = skyline_matrix(
        diversification_between_types, N=N, zero_diagonal=True
    )

    birth_rates = diversification / (1 - turnover)
    death_rates = turnover * birth_rates
    sampling_rates = (
        sampling_proportions
        * death_rates
        / (1 - removal_probabilities * sampling_proportions)
    )
    birth_rates_among_demes = diversification_between_types + death_rates

    return get_canonical_reactions(
        populations=populations,
        sample_population=sample_population,
        birth_rates=birth_rates,
        death_rates=death_rates,
        sampling_rates=sampling_rates,
        removal_probabilities=removal_probabilities,
        migration_rates=migration_rates,
        birth_rates_among_demes=birth_rates_among_demes,
    )



from phylogenie.skyline.matrix import (
    SkylineMatrix,
    SkylineMatrixCoercible,
    skyline_matrix,
)
from phylogenie.skyline.parameter import (
    SkylineParameter,
    SkylineParameterLike,
    skyline_parameter,
)
from phylogenie.skyline.vector import (
    SkylineVector,
    SkylineVectorCoercible,
    SkylineVectorLike,
    skyline_vector,
)

__all__ = [
    "skyline_matrix",
    "skyline_parameter",
    "skyline_vector",
    "SkylineMatrix",
    "SkylineMatrixCoercible",
    "SkylineParameter",
    "SkylineParameterLike",
    "SkylineVector",
    "SkylineVectorCoercible",
    "SkylineVectorLike",
]



from abc import ABC, abstractmethod
from enum import Enum
from typing import Annotated, Generic, Literal, TypeVar

from numpy.random import Generator, default_rng
from pydantic import Field

import phylogenie.typings as pgt
from phylogenie.configs import StrictBaseModel

_T = TypeVar("_T")


class Type(str, Enum):
    UNIFORM = "uniform"
    NORMAL = "normal"
    LOGNORMAL = "lognormal"
    WEIBULL = "weibull"
    EXPONENTIAL = "exponential"
    GAMMA = "gamma"
    BETA = "beta"
    INT_UNIFORM = "int-uniform"
    CATEGORICAL = "categorical"


class Distribution(StrictBaseModel, ABC, Generic[_T]):
    @abstractmethod
    def _sample(self, rng: Generator) -> _T: ...

    def sample(self, rng: int | Generator | None = None) -> _T:
        if not isinstance(rng, Generator):
            rng = default_rng(rng)
        return self._sample(rng)


class Scalar(Distribution[pgt.Scalar]): ...


class Uniform(Scalar):
    type: Literal[Type.UNIFORM] = Type.UNIFORM
    low: float
    high: float

    def _sample(self, rng: Generator) -> float:
        return rng.uniform(self.low, self.high)


class Normal(Scalar):
    type: Literal[Type.NORMAL] = Type.NORMAL
    mean: float
    std: float

    def _sample(self, rng: Generator) -> float:
        return rng.normal(self.mean, self.std)


class LogNormal(Scalar):
    type: Literal[Type.LOGNORMAL] = Type.LOGNORMAL
    mean: float
    std: float

    def _sample(self, rng: Generator) -> float:
        return rng.lognormal(self.mean, self.std)


class Weibull(Scalar):
    type: Literal[Type.WEIBULL] = Type.WEIBULL
    scale: float
    shape: float

    def _sample(self, rng: Generator) -> float:
        return rng.weibull(self.shape) * self.scale


class Exponential(Scalar):
    type: Literal[Type.EXPONENTIAL] = Type.EXPONENTIAL
    scale: float

    def _sample(self, rng: Generator) -> float:
        return rng.exponential(scale=self.scale)


class Gamma(Scalar):
    type: Literal[Type.GAMMA] = Type.GAMMA
    scale: float
    shape: float

    def _sample(self, rng: Generator) -> float:
        return rng.gamma(shape=self.shape, scale=self.scale)


class Beta(Scalar):
    type: Literal[Type.BETA] = Type.BETA
    alpha: float
    beta: float

    def _sample(self, rng: Generator) -> float:
        return rng.beta(a=self.alpha, b=self.beta)


class IntUniform(Scalar):
    type: Literal[Type.INT_UNIFORM] = Type.INT_UNIFORM
    low: int
    high: int

    def _sample(self, rng: Generator) -> int:
        return int(rng.integers(self.low, self.high))


class Categorical(Distribution[str]):
    type: Literal[Type.CATEGORICAL] = Type.CATEGORICAL
    categories: list[str]
    probabilities: list[float]

    def _sample(self, rng: Generator) -> str:
        return str(rng.choice(self.categories, p=self.probabilities))


ScalarDistributionConfig = Annotated[
    Uniform | Normal | LogNormal | Weibull | Exponential | Gamma | Beta | IntUniform,
    Field(discriminator="type"),
]
DistributionConfig = Annotated[
    ScalarDistributionConfig | Categorical, Field(discriminator="type")
]



import os
from argparse import ArgumentParser
from glob import glob

from pydantic import TypeAdapter
from yaml import safe_load

from phylogenie.core import DatasetGeneratorConfig
from phylogenie.core.dataset import DatasetGenerator


def run(config_path: str) -> None:
    adapter: TypeAdapter[DatasetGenerator] = TypeAdapter(DatasetGeneratorConfig)

    if os.path.isdir(config_path):
        for config_file in glob(os.path.join(config_path, "**/*.yaml"), recursive=True):
            with open(config_file, "r") as f:
                config = safe_load(f)
            generator = adapter.validate_python(config)
            generator.generate()
    else:
        with open(config_path, "r") as f:
            config = safe_load(f)
        generator = adapter.validate_python(config)
        generator.generate()


def main() -> None:
    parser = ArgumentParser(
        description="Generate dataset(s) starting from provided config(s)."
    )
    parser.add_argument(
        "config_path",
        type=str,
        help="Path to a config file or a directory containing config files.",
    )
    args = parser.parse_args()

    run(args.config_path)



from collections.abc import Sequence
from typing import TypeGuard

import phylogenie.typings as pgt


def is_many(x: object) -> TypeGuard[pgt.Many[object]]:
    return isinstance(x, Sequence) and not isinstance(x, str)


def is_many_scalars(x: object) -> TypeGuard[pgt.Many[pgt.Scalar]]:
    return is_many(x) and all(isinstance(i, pgt.Scalar) for i in x)


def is_many_ints(x: object) -> TypeGuard[pgt.Many[int]]:
    return is_many(x) and all(isinstance(i, int) for i in x)


def is_one_or_many_scalars(x: object) -> TypeGuard[pgt.OneOrManyScalars]:
    return isinstance(x, pgt.Scalar) or is_many_scalars(x)


def is_many_one_or_many_scalars(x: object) -> TypeGuard[pgt.Many[pgt.OneOrManyScalars]]:
    return is_many(x) and all(is_one_or_many_scalars(i) for i in x)


def is_many_2D_scalars(x: object) -> TypeGuard[pgt.Many2DScalars]:
    return is_many(x) and all(is_many_scalars(i) for i in x)


def is_one_or_many_2D_scalars(x: object) -> TypeGuard[pgt.OneOrMany2DScalars]:
    return isinstance(x, pgt.Scalar) or is_many_2D_scalars(x)


def is_many_one_or_many_2D_scalars(
    x: object,
) -> TypeGuard[pgt.Many[pgt.OneOrMany2DScalars]]:
    return is_many(x) and all(is_one_or_many_2D_scalars(i) for i in x)


def is_many_3D_scalars(x: object) -> TypeGuard[pgt.Many3DScalars]:
    return is_many(x) and all(is_many_2D_scalars(i) for i in x)



from dataclasses import dataclass

import numpy as np
from treesimulator import STATE, save_forest
from treesimulator.generator import generate
from treesimulator.mtbd_models import CTModel, Model

from phylogenie.skyline import (
    SkylineMatrixCoercible,
    SkylineParameterLike,
    SkylineVectorCoercible,
    skyline_matrix,
    skyline_parameter,
    skyline_vector,
)

DEFAULT_POPULATION = "X"
INFECTIOUS_POPULATION = "I"
EXPOSED_POPULATION = "E"


@dataclass
class TreeParams:
    populations: str | list[str] = DEFAULT_POPULATION
    transition_rates: SkylineMatrixCoercible = 0
    transmission_rates: SkylineMatrixCoercible = 0
    removal_rates: SkylineVectorCoercible = 0
    sampling_proportions: SkylineVectorCoercible = 0


def generate_tree(
    output_file: str,
    params: TreeParams,
    min_tips: int,
    max_tips: int,
    T: float = np.inf,
    state_frequencies: list[float] | None = None,
    notification_probability: SkylineParameterLike = 0,
    notification_sampling_rate: SkylineParameterLike = np.inf,
    allow_irremovable_states: bool = False,
    max_notified_contacts: int = 1,
    root_state: str | None = None,
    random_seed: int | None = None,
) -> None:
    populations = params.populations
    if isinstance(populations, str):
        populations = [populations]
    N = len(populations)

    transition_rates = skyline_matrix(params.transition_rates, N=N, zero_diagonal=True)
    transmission_rates = skyline_matrix(params.transmission_rates, N=N)
    removal_rates = skyline_vector(params.removal_rates, N=N)
    sampling_proportions = skyline_vector(params.sampling_proportions, N=N)

    change_times = sorted(
        set(
            [
                *transition_rates.change_times,
                *transmission_rates.change_times,
                *removal_rates.change_times,
                *sampling_proportions.change_times,
            ]
        )
    )

    models = [
        Model(
            states=populations,
            transition_rates=transition_rates.get_value_at_time(t),
            transmission_rates=transmission_rates.get_value_at_time(t),
            removal_rates=removal_rates.get_value_at_time(t),
            ps=sampling_proportions.get_value_at_time(t),
        )
        for t in [0, *change_times]
    ]

    if notification_probability:
        notification_sampling_rate = skyline_parameter(notification_sampling_rate)
        notification_probability = skyline_parameter(notification_probability)
        models = [
            CTModel(
                model,
                phi=notification_sampling_rate.get_value_at_time(t),
                upsilon=notification_probability.get_value_at_time(t),
                allow_irremovable_states=allow_irremovable_states,
            )
            for t, model in zip([0, *change_times], models)
        ]

    [tree], _, _ = generate(
        models,
        min_tips=min_tips,
        max_tips=max_tips,
        T=T,
        skyline_times=change_times,
        state_frequencies=state_frequencies,
        max_notified_contacts=max_notified_contacts,
        root_state=root_state,
        random_seed=random_seed,
    )
    for i, leaf in enumerate(tree.iter_leaves()):
        state: str = getattr(leaf, STATE)
        date = tree.get_distance(leaf)
        leaf.name = f"{i}|{state}|{date}"
    save_forest([tree], output_file)


def get_BD_params(
    reproduction_number: SkylineParameterLike,
    infectious_period: SkylineParameterLike,
    sampling_proportion: SkylineParameterLike,
) -> TreeParams:
    transmission_rate = reproduction_number / infectious_period
    removal_rate = 1 / infectious_period
    return TreeParams(
        populations=INFECTIOUS_POPULATION,
        transmission_rates=transmission_rate,
        removal_rates=removal_rate,
        sampling_proportions=sampling_proportion,
    )


def get_BDEI_params(
    reproduction_number: SkylineParameterLike,
    infectious_period: SkylineParameterLike,
    incubation_period: SkylineParameterLike,
    sampling_proportion: SkylineParameterLike,
) -> TreeParams:
    transmission_rates = [[0, 0], [reproduction_number / infectious_period, 0]]
    transition_rates = [[0, 1 / incubation_period], [0, 0]]
    removal_rates = [0, 1 / infectious_period]
    sampling_proportions = [0, sampling_proportion]
    return TreeParams(
        populations=[EXPOSED_POPULATION, INFECTIOUS_POPULATION],
        transition_rates=transition_rates,
        transmission_rates=transmission_rates,
        removal_rates=removal_rates,
        sampling_proportions=sampling_proportions,
    )



from collections.abc import Callable, Iterator
from typing import TypeGuard, Union

import phylogenie.typeguards as tg
import phylogenie.typings as pgt
from phylogenie.skyline.parameter import (
    SkylineParameter,
    SkylineParameterLike,
    is_many_skyline_parameters_like,
    is_skyline_parameter_like,
    skyline_parameter,
)

SkylineVectorOperand = Union[SkylineParameterLike, "SkylineVector"]
SkylineVectorLike = Union[pgt.Many[SkylineParameterLike], "SkylineVector"]
SkylineVectorCoercible = Union[pgt.OneOrMany[SkylineParameterLike], "SkylineVector"]


def is_skyline_vector_operand(x: object) -> TypeGuard[SkylineVectorOperand]:
    return isinstance(x, SkylineVector) or is_skyline_parameter_like(x)


def is_skyline_vector_like(x: object) -> TypeGuard[SkylineVectorLike]:
    return isinstance(x, SkylineVector) or is_many_skyline_parameters_like(x)


def is_many_skyline_vectors_like(x: object) -> TypeGuard[pgt.Many[SkylineVectorLike]]:
    return tg.is_many(x) and all(is_skyline_vector_like(v) for v in x)


class SkylineVector:
    def __init__(
        self,
        params: pgt.Many[SkylineParameterLike] | None = None,
        value: pgt.Many2DScalars | None = None,
        change_times: pgt.ManyScalars | None = None,
    ) -> None:
        if params is not None and value is None and change_times is None:
            if is_many_skyline_parameters_like(params):
                self.params = [skyline_parameter(param) for param in params]
            else:
                raise TypeError(
                    f"It is impossible to create a SkylineVector from `params` {params} of type {type(params)}. Please provide a sequence of SkylineParameterLike objects (a SkylineParameterLike object can either be a SkylineParameter or a scalar)."
                )
        elif value is not None and change_times is not None:
            if tg.is_many_2D_scalars(value):
                vector_lengths = {len(vector) for vector in value}
                if any(vl != len(value[0]) for vl in vector_lengths):
                    raise ValueError(
                        f"All rows in the `value` of a SkylineVector must have the same length (got value={value} with vector lengths={vector_lengths})."
                    )
            else:
                raise TypeError(
                    f"It is impossible to create a SkylineVector from `value` {value} of type {type(value)}. Please provide a nested (2D) sequence of scalar values."
                )
            self.params = [
                SkylineParameter([vector[i] for vector in value], change_times)
                for i in range(len(value[0]))
            ]
        else:
            raise ValueError(
                "Either `params` or both `value` and `change_times` must be provided to create a SkylineVector."
            )

    @property
    def change_times(self) -> pgt.Vector1D:
        return sorted(set(t for param in self.params for t in param.change_times))

    @property
    def value(self) -> pgt.Vector2D:
        return [self.get_value_at_time(t) for t in (0, *self.change_times)]

    @property
    def N(self) -> int:
        return len(self.params)

    def get_value_at_time(self, t: pgt.Scalar) -> pgt.Vector1D:
        return [param.get_value_at_time(t) for param in self.params]

    def operate(
        self,
        other: SkylineVectorOperand,
        func: Callable[[SkylineParameter, SkylineParameter], SkylineParameter],
    ) -> "SkylineVector":
        if not is_skyline_vector_operand(other):
            return NotImplemented
        other = skyline_vector(other, self.N)
        return SkylineVector(
            [func(p1, p2) for p1, p2 in zip(self.params, other.params)]
        )

    def __add__(self, operand: SkylineVectorOperand) -> "SkylineVector":
        return self.operate(operand, lambda x, y: x + y)

    def __radd__(self, operand: SkylineParameterLike) -> "SkylineVector":
        return self.operate(operand, lambda x, y: y + x)

    def __sub__(self, operand: SkylineVectorOperand) -> "SkylineVector":
        return self.operate(operand, lambda x, y: x - y)

    def __rsub__(self, operand: SkylineParameterLike) -> "SkylineVector":
        return self.operate(operand, lambda x, y: y - x)

    def __mul__(self, operand: SkylineVectorOperand) -> "SkylineVector":
        return self.operate(operand, lambda x, y: x * y)

    def __rmul__(self, operand: SkylineParameterLike) -> "SkylineVector":
        return self.operate(operand, lambda x, y: y * x)

    def __truediv__(self, operand: SkylineVectorOperand) -> "SkylineVector":
        return self.operate(operand, lambda x, y: x / y)

    def __rtruediv__(self, operand: SkylineParameterLike) -> "SkylineVector":
        return self.operate(operand, lambda x, y: y / x)

    def __len__(self) -> int:
        return self.N

    def __bool__(self) -> bool:
        return any(self.params)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SkylineVector) and self.params == other.params

    def __repr__(self) -> str:
        return f"SkylineVector(value={list(self.value)}, change_times={list(self.change_times)})"

    def __iter__(self) -> Iterator[SkylineParameter]:
        return iter(self.params)

    def __getitem__(self, item: int) -> "SkylineParameter":
        return self.params[item]

    def __setitem__(self, item: int, value: SkylineParameterLike) -> None:
        if not is_skyline_parameter_like(value):
            raise TypeError(
                f"It is impossible to set item {item} of SkylineVector with value {value} of type {type(value)}. Please provide a SkylineParameterLike object (i.e., a scalar or a SkylineParameter)."
            )
        self.params[item] = skyline_parameter(value)


def skyline_vector(x: SkylineVectorCoercible, N: int) -> SkylineVector:
    if is_skyline_parameter_like(x):
        return SkylineVector([skyline_parameter(x)] * N)
    elif is_many_skyline_parameters_like(x):
        x = SkylineVector(x)
    elif not isinstance(x, SkylineVector):
        raise TypeError(
            f"It is impossible to coerce {x} of type {type(x)} into a SkylineVector. Please provide a SkylineParameterLike object (i.e., a scalar or a SkylineParameter), or a sequence of them."
        )
    if x.N != N:
        raise ValueError(
            f"Expected a SkylineVector of size {N}, got {x} of size {x.N}."
        )
    return x



import phylogenie.typings as pgt
from phylogenie.configs import StrictBaseModel

IntConfig = str | int
ScalarConfig = str | pgt.Scalar
ManyIntsConfig = str | list[IntConfig]
ManyScalarsConfig = str | list[ScalarConfig]
OneOrManyScalarsConfig = ScalarConfig | list[ScalarConfig]
OneOrMany2DScalarsConfig = ScalarConfig | list[list[ScalarConfig]]


class SkylineParameterValueModel(StrictBaseModel):
    value: ManyScalarsConfig
    change_times: ManyScalarsConfig


SkylineParameterLikeConfig = ScalarConfig | SkylineParameterValueModel


class SkylineVectorValueModel(StrictBaseModel):
    value: str | list[OneOrManyScalarsConfig]
    change_times: ManyScalarsConfig


SkylineVectorCoercibleConfig = (
    str | pgt.Scalar | list[SkylineParameterLikeConfig] | SkylineVectorValueModel
)
SkylineVectorLikeConfig = (
    str | list[SkylineParameterLikeConfig] | SkylineVectorValueModel
)


class SkylineMatrixValueModel(StrictBaseModel):
    value: str | list[OneOrMany2DScalarsConfig]
    change_times: ManyScalarsConfig


SkylineMatrixCoercibleConfig = (
    str | pgt.Scalar | list[SkylineVectorLikeConfig] | SkylineMatrixValueModel
)



import os
from abc import abstractmethod
from enum import Enum
from pathlib import Path
from typing import Literal

from numpy.random import Generator

import phylogenie.typings as pgt
from phylogenie.core.dataset import DatasetGenerator, DataType
from phylogenie.core.trees import TreesGeneratorConfig


class BackendType(str, Enum):
    ALISIM = "alisim"


class MSAsGenerator(DatasetGenerator):
    data_type: Literal[DataType.MSAS] = DataType.MSAS
    trees: str | TreesGeneratorConfig
    output_trees_dir: str | None = None

    @abstractmethod
    def _generate_one_from_tree(
        self, filename: str, tree_file: str, rng: Generator, data: pgt.Data
    ) -> None: ...

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        if isinstance(self.trees, str):
            tree_files = os.listdir(self.trees)
            tree_file = os.path.join(self.trees, str(rng.choice(tree_files)))
            self._generate_one_from_tree(
                filename=filename, tree_file=tree_file, rng=rng, data=data
            )
        elif isinstance(self.output_trees_dir, str):
            os.makedirs(self.output_trees_dir, exist_ok=True)
            self._generate_one_from_tree(
                filename=filename,
                tree_file=os.path.join(self.output_trees_dir, f"{Path(filename).stem}"),
                rng=rng,
                data=data,
            )
        else:
            tree_filename = f"{filename}.temp-tree"
            self.trees.generate_one(
                filename=tree_filename, data=data, seed=int(rng.integers(0, 2**32 - 1))
            )
            self._generate_one_from_tree(
                filename=filename, tree_file=f"{tree_filename}.nwk", rng=rng, data=data
            )
            os.remove(f"{tree_filename}.nwk")



import os
from abc import ABC, abstractmethod
from enum import Enum

import joblib
import pandas as pd
from numpy.random import Generator, default_rng
from tqdm import tqdm

import phylogenie.typings as pgt
from phylogenie.configs import StrictBaseModel
from phylogenie.core.context import ContextConfig, context_factory


class DataType(str, Enum):
    TREES = "trees"
    MSAS = "msas"


class DatasetGenerator(ABC, StrictBaseModel):
    output_dir: str = "phylogenie-out"
    data_dir: str = "data"
    metadata_filename: str = "metadata.csv"
    n_samples: int | dict[str, int] = 1
    n_jobs: int = -1
    seed: int | None = None
    context: ContextConfig | None = None

    @abstractmethod
    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None: ...

    def generate_one(
        self, filename: str, data: pgt.Data | None = None, seed: int | None = None
    ) -> None:
        data = {} if data is None else data
        self._generate_one(filename=filename, rng=default_rng(seed), data=data)

    def _generate(self, rng: Generator, n_samples: int, output_dir: str) -> None:
        data_dir = os.path.join(output_dir, self.data_dir)
        metadata_file = os.path.join(output_dir, self.metadata_filename)
        if os.path.exists(data_dir):
            print(f"Output directory {data_dir} already exists. Skipping.")
            return
        os.makedirs(data_dir)

        data = [
            {} if self.context is None else context_factory(self.context, rng)
            for _ in range(n_samples)
        ]

        joblib.Parallel(n_jobs=self.n_jobs)(
            joblib.delayed(self.generate_one)(
                filename=os.path.join(data_dir, str(i)),
                data=d,
                seed=int(rng.integers(0, 2**32)),
            )
            for i, d in tqdm(
                enumerate(data), total=n_samples, desc=f"Generating {data_dir}..."
            )
        )

        df = pd.DataFrame([{"file_id": str(i), **d} for i, d in enumerate(data)])
        df.to_csv(metadata_file, index=False)

    def generate(self) -> None:
        rng = default_rng(self.seed)
        if isinstance(self.n_samples, dict):
            for key, n_samples in self.n_samples.items():
                output_dir = os.path.join(self.output_dir, key)
                self._generate(rng, n_samples, output_dir)
        else:
            self._generate(rng, self.n_samples, self.output_dir)



from numpy.random import Generator

import phylogenie.core.context.configs as cfg
import phylogenie.typings as pgt
from phylogenie.core.context import distributions


def _sample_vector1D(x: distributions.Scalar, N: int, rng: Generator) -> pgt.Vector1D:
    return [x.sample(rng) for _ in range(N)]


def _sample_vector2D(
    x: distributions.Scalar, N: int, zero_diagonal: bool, rng: Generator
) -> pgt.Vector2D:
    v = [_sample_vector1D(x, N, rng) for _ in range(N)]
    if zero_diagonal:
        for i in range(N):
            v[i][i] = 0
    return v


def _sample_vector3D(
    x: distributions.Scalar, N: int, T: int, zero_diagonal: bool, rng: Generator
) -> pgt.Vector3D:
    return [_sample_vector2D(x, N, zero_diagonal, rng) for _ in range(T)]


def context_factory(x: cfg.ContextConfig, rng: Generator) -> pgt.Data:
    data: pgt.Data = {}
    for key, value in x.items():
        if isinstance(value, distributions.Distribution):
            data[key] = value.sample(rng)
        elif isinstance(value, cfg.Vector1DModel):
            data[key] = _sample_vector1D(value.x, value.N, rng)
        elif isinstance(value, cfg.Vector2DModel):
            data[key] = _sample_vector2D(value.x, value.N, value.zero_diagonal, rng)
        else:
            data[key] = _sample_vector3D(
                value.x, value.N, value.T, value.zero_diagonal, rng
            )
    return data



from typing import Annotated, Literal

from pydantic import Field

from phylogenie.configs import StrictBaseModel
from phylogenie.core.context.distributions import (
    DistributionConfig,
    ScalarDistributionConfig,
)


class VectorModel(StrictBaseModel):
    x: ScalarDistributionConfig
    N: int


class Vector1DModel(VectorModel):
    n_dim: Literal[1] = 1


class Vector2DModel(VectorModel):
    n_dim: Literal[2] = 2
    zero_diagonal: bool = False


class Vector3DModel(VectorModel):
    n_dim: Literal[3] = 3
    T: int
    zero_diagonal: bool = False


ContextConfig = dict[
    str,
    DistributionConfig
    | Annotated[
        Vector1DModel | Vector2DModel | Vector3DModel, Field(discriminator="n_dim")
    ],
]



from typing import TypeGuard

import phylogenie.core.configs as cfg


def is_list(x: object) -> TypeGuard[list[object]]:
    return isinstance(x, list)


def is_list_of_scalar_configs(x: object) -> TypeGuard[list[cfg.ScalarConfig]]:
    return is_list(x) and all(isinstance(v, cfg.ScalarConfig) for v in x)


def is_list_of_skyline_parameter_like_configs(
    x: object,
) -> TypeGuard[list[cfg.SkylineParameterLikeConfig]]:
    return is_list(x) and all(isinstance(v, cfg.SkylineParameterLikeConfig) for v in x)


def is_list_of_skyline_vector_like_configs(
    x: object,
) -> TypeGuard[list[cfg.SkylineVectorLikeConfig]]:
    return is_list(x) and all(
        isinstance(v, str | cfg.SkylineVectorValueModel)
        or is_list_of_skyline_parameter_like_configs(v)
        for v in x
    )



from collections.abc import Iterable
from enum import Enum
from typing import Annotated, Literal

from numpy.random import Generator
from pydantic import Field

import phylogenie.core.configs as cfg
import phylogenie.typings as pgt
from phylogenie.backend.remaster import (
    DEFAULT_POPULATION,
    SAMPLE_POPULATION,
    Reaction,
    generate_trees,
    get_canonical_reactions,
    get_epidemiological_reactions,
    get_FBD_reactions,
)
from phylogenie.core.factories import (
    skyline_matrix_coercible_factory,
    skyline_vector_coercible_factory,
)
from phylogenie.core.trees.base import BackendType, TreesGenerator
from phylogenie.core.trees.remaster.configs import (
    PunctualReactionConfig,
    ReactionConfig,
)
from phylogenie.core.trees.remaster.factories import (
    punctual_reaction_factory,
    reaction_factory,
)


class ParameterizationType(str, Enum):
    CANONICAL = "canonical"
    EPIDEMIOLOGICAL = "epidemiological"
    FBD = "fbd"


class ReMASTERGenerator(TreesGenerator):
    backend: Literal[BackendType.REMASTER] = BackendType.REMASTER
    beast_path: str = "beast"
    populations: str | list[str] = DEFAULT_POPULATION
    init_population: str = DEFAULT_POPULATION
    sample_population: str = SAMPLE_POPULATION
    reactions: Iterable[ReactionConfig] = Field(default_factory=tuple)
    punctual_reactions: Iterable[PunctualReactionConfig] = Field(default_factory=tuple)
    trajectory_attrs: dict[str, str | int | float] = Field(default_factory=dict)

    def _generate_one_from_extra_reactions(
        self, filename: str, rng: Generator, data: pgt.Data, reactions: list[Reaction]
    ) -> None:
        generate_trees(
            tree_file_name=f"{filename}.nwk",
            populations=self.populations,
            init_population=self.init_population.format(**data),
            sample_population=self.sample_population,
            reactions=[reaction_factory(r, data) for r in self.reactions] + reactions,
            punctual_reactions=[
                punctual_reaction_factory(r, data) for r in self.punctual_reactions
            ],
            trajectory_attrs={
                k: v.format(**data) if isinstance(v, str) else str(v)
                for k, v in self.trajectory_attrs.items()
            },
            seed=int(rng.integers(0, 2**31 - 1)),
            beast_path=self.beast_path,
        )

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        self._generate_one_from_extra_reactions(filename, rng, data, reactions=[])


class CanonicalReMASTERGenerator(ReMASTERGenerator):
    parameterization: Literal[ParameterizationType.CANONICAL] = (
        ParameterizationType.CANONICAL
    )
    birth_rates: cfg.SkylineVectorCoercibleConfig = 0
    death_rates: cfg.SkylineVectorCoercibleConfig = 0
    sampling_rates: cfg.SkylineVectorCoercibleConfig = 0
    removal_probabilities: cfg.SkylineVectorCoercibleConfig = 0
    migration_rates: cfg.SkylineMatrixCoercibleConfig = 0
    birth_rates_among_demes: cfg.SkylineMatrixCoercibleConfig = 0

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        reactions = get_canonical_reactions(
            populations=self.populations,
            sample_population=self.sample_population,
            birth_rates=skyline_vector_coercible_factory(self.birth_rates, data),
            death_rates=skyline_vector_coercible_factory(self.death_rates, data),
            sampling_rates=skyline_vector_coercible_factory(self.sampling_rates, data),
            removal_probabilities=skyline_vector_coercible_factory(
                self.removal_probabilities, data
            ),
            migration_rates=skyline_matrix_coercible_factory(
                self.migration_rates, data
            ),
            birth_rates_among_demes=skyline_matrix_coercible_factory(
                self.birth_rates_among_demes, data
            ),
        )
        self._generate_one_from_extra_reactions(filename, rng, data, reactions)


class EpidemiologicalReMASTERGenerator(ReMASTERGenerator):
    parameterization: Literal[ParameterizationType.EPIDEMIOLOGICAL] = (
        ParameterizationType.EPIDEMIOLOGICAL
    )
    reproduction_numbers: cfg.SkylineVectorCoercibleConfig = 0
    become_uninfectious_rates: cfg.SkylineVectorCoercibleConfig = 0
    sampling_proportions: cfg.SkylineVectorCoercibleConfig = 0
    removal_probabilities: cfg.SkylineVectorCoercibleConfig = 0
    migration_rates: cfg.SkylineMatrixCoercibleConfig = 0
    reproduction_numbers_among_demes: cfg.SkylineMatrixCoercibleConfig = 0

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        reactions = get_epidemiological_reactions(
            populations=self.populations,
            sample_population=self.sample_population,
            reproduction_numbers=skyline_vector_coercible_factory(
                self.reproduction_numbers, data
            ),
            become_uninfectious_rates=skyline_vector_coercible_factory(
                self.become_uninfectious_rates, data
            ),
            sampling_proportions=skyline_vector_coercible_factory(
                self.sampling_proportions, data
            ),
            removal_probabilities=skyline_vector_coercible_factory(
                self.removal_probabilities, data
            ),
            migration_rates=skyline_matrix_coercible_factory(
                self.migration_rates, data
            ),
            reproduction_numbers_among_demes=skyline_matrix_coercible_factory(
                self.reproduction_numbers_among_demes, data
            ),
        )
        self._generate_one_from_extra_reactions(filename, rng, data, reactions)


class FBDReMASTERGenerator(ReMASTERGenerator):
    parameterization: Literal[ParameterizationType.FBD] = ParameterizationType.FBD
    diversification: cfg.SkylineVectorCoercibleConfig = 0
    turnover: cfg.SkylineVectorCoercibleConfig = 0
    sampling_proportions: cfg.SkylineVectorCoercibleConfig = 0
    removal_probabilities: cfg.SkylineVectorCoercibleConfig = 0
    migration_rates: cfg.SkylineMatrixCoercibleConfig = 0
    diversification_between_types: cfg.SkylineMatrixCoercibleConfig = 0

    def _generate_one(self, filename: str, rng: Generator, data: pgt.Data) -> None:
        reactions = get_FBD_reactions(
            populations=self.populations,
            sample_population=self.sample_population,
            diversification=skyline_vector_coercible_factory(
                self.diversification, data
            ),
            turnover=skyline_vector_coercible_factory(self.turnover, data),
            sampling_proportions=skyline_vector_coercible_factory(
                self.sampling_proportions, data
            ),
            removal_probabilities=skyline_vector_coercible_factory(
                self.removal_probabilities, data
            ),
            migration_rates=skyline_matrix_coercible_factory(
                self.migration_rates, data
            ),
            diversification_between_types=skyline_matrix_coercible_factory(
                self.diversification_between_types, data
            ),
        )
        self._generate_one_from_extra_reactions(filename, rng, data, reactions)


ReMASTERGeneratorConfig = Annotated[
    CanonicalReMASTERGenerator
    | EpidemiologicalReMASTERGenerator
    | FBDReMASTERGenerator,
    Field(discriminator="parameterization"),
]



import phylogenie.core.configs as cfg
from phylogenie.configs import StrictBaseModel


class ReactionConfig(StrictBaseModel):
    rate: cfg.SkylineParameterLikeConfig
    value: str


class PunctualReactionConfig(StrictBaseModel):
    times: cfg.ManyScalarsConfig
    value: str
    p: cfg.ManyScalarsConfig | None = None
    n: cfg.ManyIntsConfig | None = None



from phylogenie.backend.remaster.generate import generate_trees
from phylogenie.backend.remaster.reactions import (
    DEFAULT_POPULATION,
    SAMPLE_POPULATION,
    PunctualReaction,
    Reaction,
    get_canonical_reactions,
    get_epidemiological_reactions,
    get_FBD_reactions,
)

__all__ = [
    "DEFAULT_POPULATION",
    "SAMPLE_POPULATION",
    "PunctualReaction",
    "Reaction",
    "get_canonical_reactions",
    "get_epidemiological_reactions",
    "get_FBD_reactions",
    "generate_trees",
]



import re
import subprocess
from collections.abc import Iterable
from xml.dom import minidom
from xml.etree.ElementTree import Element, tostring

from phylogenie.backend.remaster.reactions import (
    DEFAULT_POPULATION,
    SAMPLE_POPULATION,
    PunctualReaction,
    Reaction,
)
from phylogenie.skyline import skyline_parameter

TREE_ID = "Tree"


def _beautify_xml(xml: bytes) -> str:
    xml_str = minidom.parseString(xml).toprettyxml(indent="\t")
    xml_str = re.sub(r"\n\s*\n+", "\n", xml_str)
    xml_str = xml_str.strip()
    return xml_str


def _generate_config_file(
    output_xml_file: str,
    tree_file_name: str,
    populations: list[str],
    init_values: list[int],
    sample_population: str,
    reactions: Iterable[Reaction] | None = None,
    punctual_reactions: Iterable[PunctualReaction] | None = None,
    trajectory_attrs: dict[str, str] | None = None,
    n_simulations: int = 1,
) -> None:
    simulate = Element("simulate", {"spec": "SimulatedTree", "id": TREE_ID})

    if trajectory_attrs is None:
        trajectory_attrs = {}
    trajectory = Element(
        "trajectory", {"spec": "StochasticTrajectory", **trajectory_attrs}
    )

    for population, init_value in zip(populations, init_values):
        trajectory.append(
            Element(
                "population",
                {"spec": "RealParameter", "id": population, "value": str(init_value)},
            )
        )
    trajectory.append(
        Element(
            "samplePopulation",
            {"spec": "RealParameter", "id": sample_population, "value": "0"},
        )
    )

    if reactions is not None:
        for reaction in reactions:
            if not reaction.rate:
                continue
            rate = skyline_parameter(reaction.rate)
            attrs = {
                "spec": "Reaction",
                "rate": " ".join(map(str, rate.value)),
                "value": reaction.value,
            }
            if rate.change_times:
                attrs["changeTimes"] = " ".join(map(str, rate.change_times))
            trajectory.append(Element("reaction", attrs))

    if punctual_reactions is not None:
        for punctual_reaction in punctual_reactions:
            attrs = {
                "spec": "PunctualReaction",
                "value": punctual_reaction.value,
                "times": " ".join(map(str, punctual_reaction.times)),
            }
            if punctual_reaction.p is not None:
                attrs["p"] = " ".join(map(str, punctual_reaction.p))
            if punctual_reaction.n is not None:
                attrs["n"] = " ".join(map(str, punctual_reaction.n))
            trajectory.append(Element("reaction", attrs))

    simulate.append(trajectory)

    logger = Element(
        "logger", {"spec": "Logger", "mode": "tree", "fileName": tree_file_name}
    )
    logger.append(
        Element(
            "log",
            {
                "spec": "TypedTreeLogger",
                "tree": f"@{TREE_ID}",
                "removeSingletonNodes": "true",
                "noLabels": "true",
            },
        )
    )

    run = Element("run", {"spec": "Simulator", "nSims": str(n_simulations)})
    run.append(simulate)
    run.append(logger)

    beast = Element(
        "beast",
        {
            "version": "2.0",
            "namespace": ":".join(
                ["beast.base.inference", "beast.base.inference.parameter", "remaster"]
            ),
        },
    )
    beast.append(run)

    with open(output_xml_file, "w") as f:
        f.write(_beautify_xml(tostring(beast, method="xml")))


def _postprocess_tree(input_file: str, output_file: str, attributes: list[str]) -> None:

    def _replace_metadata(match: re.Match[str]) -> str:
        metadata = match.group(0)
        attrs: list[tuple[str, str]] = re.findall(r'(\w+)=(".*?"|[^,)\]]+)', metadata)
        values = [v.strip('"') for k, v in attrs if k in attributes]
        return "|" + "|".join(values)

    with open(input_file, "r") as infile:
        with open(output_file, "w") as outfile:
            for line in infile:
                line = line.strip()
                if line.lower().startswith("tree"):
                    parts = line.split("=", 1)
                    newick = parts[1].strip()
                    transformed_newick = re.sub(
                        r"\[\&[^\]]*\]", _replace_metadata, newick
                    )
                    outfile.write(transformed_newick + "\n")


def generate_trees(
    tree_file_name: str,
    populations: str | list[str] = DEFAULT_POPULATION,
    init_population: str = DEFAULT_POPULATION,
    sample_population: str = SAMPLE_POPULATION,
    reactions: Iterable[Reaction] | None = None,
    punctual_reactions: Iterable[PunctualReaction] | None = None,
    trajectory_attrs: dict[str, str] | None = None,
    output_xml_file: str | None = None,
    n_simulations: int = 1,
    seed: int | None = None,
    beast_path: str = "beast",
) -> None:
    if isinstance(populations, str):
        populations = [populations]
    init_values = [0] * len(populations)
    init_values[populations.index(init_population)] = 1

    if output_xml_file is None:
        xml_file = f"{tree_file_name}-temp.xml"
    else:
        xml_file = output_xml_file

    temp_tree_file = f"{tree_file_name}-temp.nex"
    _generate_config_file(
        output_xml_file=xml_file,
        tree_file_name=temp_tree_file,
        populations=populations,
        init_values=init_values,
        sample_population=sample_population,
        reactions=reactions,
        punctual_reactions=punctual_reactions,
        trajectory_attrs=trajectory_attrs,
        n_simulations=n_simulations,
    )

    cmd = [beast_path]
    if seed is not None:
        cmd.extend(["-seed", str(seed)])
    cmd.append(xml_file)
    subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    _postprocess_tree(temp_tree_file, tree_file_name, ["type", "time"])
    if output_xml_file is None:
        subprocess.run(["rm", xml_file], check=True)
    subprocess.run(["rm", temp_tree_file], check=True)



from enum import Enum
from typing import Literal

from phylogenie.core.dataset import DatasetGenerator, DataType


class BackendType(str, Enum):
    REMASTER = "remaster"
    TREESIMULATOR = "treesimulator"


class TreesGenerator(DatasetGenerator):
    data_type: Literal[DataType.TREES] = DataType.TREES



from typing import Any, overload

import numpy as np

import phylogenie.core.configs as cfg
import phylogenie.core.typeguards as ctg
import phylogenie.typeguards as tg
import phylogenie.typings as pgt
from phylogenie.skyline import (
    SkylineMatrix,
    SkylineMatrixCoercible,
    SkylineParameter,
    SkylineParameterLike,
    SkylineVector,
    SkylineVectorCoercible,
    SkylineVectorLike,
)


def _eval_expression(expression: str, data: pgt.Data) -> Any:
    return np.array(
        eval(
            expression,
            {
                "__builtins__": __builtins__,
                # "np": np,
                # **{k: np.array(v) for k, v in data.items()},
            },
            {k: np.array(v) for k, v in data.items()},
        )
    ).tolist()


def int_factory(x: cfg.IntConfig, data: pgt.Data) -> int:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if isinstance(e, int):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected an int."
        )
    return x


def scalar_factory(x: cfg.ScalarConfig, data: pgt.Data) -> pgt.Scalar:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if isinstance(e, pgt.Scalar):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a scalar."
        )
    return x


def many_ints_factory(x: cfg.ManyIntsConfig, data: pgt.Data) -> pgt.Many[int]:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_many_ints(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a sequence of integers."
        )
    return [int_factory(v, data) for v in x]


def many_scalars_factory(x: cfg.ManyScalarsConfig, data: pgt.Data) -> pgt.ManyScalars:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a sequence of scalars."
        )
    return [scalar_factory(v, data) for v in x]


def one_or_many_scalars_factory(
    x: cfg.OneOrManyScalarsConfig, data: pgt.Data
) -> pgt.OneOrManyScalars:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a scalar or a sequence of them."
        )
    if isinstance(x, pgt.Scalar):
        return x
    return many_scalars_factory(x, data)


def skyline_parameter_like_factory(
    x: cfg.SkylineParameterLikeConfig, data: pgt.Data
) -> SkylineParameterLike:
    if isinstance(x, cfg.ScalarConfig):
        return scalar_factory(x, data)
    return SkylineParameter(
        value=many_scalars_factory(x.value, data),
        change_times=many_scalars_factory(x.change_times, data),
    )


from typing import Literal


@overload
def _parse_skyline_vector_value_model(
    x: cfg.SkylineVectorValueModel, data: pgt.Data, coercible: Literal[True]
) -> SkylineVector | SkylineParameter: ...
@overload
def _parse_skyline_vector_value_model(
    x: cfg.SkylineVectorValueModel, data: pgt.Data, coercible: Literal[False]
) -> SkylineVector: ...
def _parse_skyline_vector_value_model(
    x: cfg.SkylineVectorValueModel, data: pgt.Data, coercible: bool
) -> SkylineVector | SkylineParameter:
    change_times = many_scalars_factory(x.change_times, data)
    if isinstance(x.value, str):
        e = _eval_expression(x.value, data)
        if tg.is_many_one_or_many_scalars(e):
            value = e
        else:
            raise ValueError(
                f"Expression '{x.value}' evaluated to {e} of type {type(e)}, which cannot be coerced to a valid value for a SkylineVector (expected a sequence composed of scalars and/or sequences of scalars)."
            )
    else:
        value = [one_or_many_scalars_factory(v, data) for v in x.value]

    if tg.is_many_scalars(value):
        if coercible:
            return SkylineParameter(value=value, change_times=change_times)
        else:
            raise ValueError(
                f"Parsing SkylineVector value {x.value} yielded a sequence of scalars ({value}) when a nested (2D) sequence of scalars was expected."
            )

    Ns = {len(elem) for elem in value if tg.is_many(elem)}
    if len(Ns) > 1:
        raise ValueError(
            f"All elements in the value of a SkylineVector config must be scalars or have the same length (got value={value} with inconsistent lengths {Ns})."
        )
    (N,) = Ns
    value = [[p] * N if isinstance(p, pgt.Scalar) else p for p in value]

    return SkylineVector(
        value=value,
        change_times=change_times,
    )


def skyline_vector_coercible_factory(
    x: cfg.SkylineVectorCoercibleConfig, data: pgt.Data
) -> SkylineVectorCoercible:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a SkylineVectorCoercible object (e.g., a scalar or a sequence of them)."
        )
    if isinstance(x, pgt.Scalar):
        return x
    if ctg.is_list_of_skyline_parameter_like_configs(x):
        return [skyline_parameter_like_factory(p, data) for p in x]

    assert isinstance(x, cfg.SkylineVectorValueModel)
    return _parse_skyline_vector_value_model(x, data, coercible=True)


def skyline_vector_like_factory(
    x: cfg.SkylineVectorLikeConfig, data: pgt.Data
) -> SkylineVectorLike:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_many_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a SkylineVectorLike object (e.g., a sequence of scalars)."
        )
    if ctg.is_list_of_skyline_parameter_like_configs(x):
        return [skyline_parameter_like_factory(p, data) for p in x]

    assert isinstance(x, cfg.SkylineVectorValueModel)
    return _parse_skyline_vector_value_model(x, data, coercible=False)


def one_or_many_2D_scalars_factory(
    x: cfg.OneOrMany2DScalarsConfig, data: pgt.Data
) -> pgt.OneOrMany2DScalars:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_2D_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a nested (2D) sequence of scalars."
        )
    if isinstance(x, pgt.Scalar):
        return x
    return [many_scalars_factory(v, data) for v in x]


def skyline_matrix_coercible_factory(
    x: cfg.SkylineMatrixCoercibleConfig, data: pgt.Data
) -> SkylineMatrixCoercible:
    if isinstance(x, str):
        e = _eval_expression(x, data)
        if tg.is_one_or_many_2D_scalars(e):
            return e
        raise ValueError(
            f"Expression '{x}' evaluated to {e} of type {type(e)}, expected a SkylineMatrixCoercible object (e.g., a scalar or a nested (2D) sequence of them)."
        )
    if isinstance(x, pgt.Scalar):
        return x
    if ctg.is_list_of_skyline_vector_like_configs(x):
        return [skyline_vector_like_factory(v, data) for v in x]

    assert isinstance(x, cfg.SkylineMatrixValueModel)

    change_times = many_scalars_factory(x.change_times, data)
    if isinstance(x.value, str):
        e = _eval_expression(x.value, data)
        if tg.is_many_one_or_many_2D_scalars(e):
            value = e
        else:
            raise ValueError(
                f"Expression '{x.value}' evaluated to {e} of type {type(e)}, which cannot be coerced to a valid value for a SkylineMatrix (expected a sequence composed of scalars and/or nested (2D) sequences of scalars)."
            )
    else:
        value = [one_or_many_2D_scalars_factory(v, data) for v in x.value]

    if tg.is_many_scalars(value):
        return SkylineParameter(value=value, change_times=change_times)

    Ns: set[int] = set()
    for elem in value:
        if tg.is_many_2D_scalars(elem):
            n_rows = len(elem)
            if any(len(row) != n_rows for row in elem):
                raise ValueError(
                    f"All elements in the value of a SkylineMatrix config must be scalars or square matrices (got non-square matrix: {elem})."
                )
            Ns.add(n_rows)

    if len(Ns) > 1:
        raise ValueError(
            f"All elements in the value of a SkylineMatrix config must be scalars or have the same square shape (got value={value} with inconsistent lengths {Ns})."
        )
    (N,) = Ns
    value = [[[p] * N] * N if isinstance(p, pgt.Scalar) else p for p in value]

    return SkylineMatrix(
        value=value,
        change_times=change_times,
    )



import phylogenie.core.trees.remaster.configs as cfg
import phylogenie.typings as pgt
from phylogenie.backend.remaster import PunctualReaction, Reaction
from phylogenie.core.factories import (
    many_ints_factory,
    many_scalars_factory,
    skyline_parameter_like_factory,
)


def reaction_factory(x: cfg.ReactionConfig, data: pgt.Data) -> Reaction:
    return Reaction(
        rate=skyline_parameter_like_factory(x.rate, data),
        value=x.value,
    )


def punctual_reaction_factory(
    x: cfg.PunctualReactionConfig, data: pgt.Data
) -> PunctualReaction:
    return PunctualReaction(
        times=many_scalars_factory(x.times, data),
        value=x.value,
        p=None if x.p is None else many_scalars_factory(x.p, data),
        n=None if x.n is None else many_ints_factory(x.n, data),
    )
